/**********************************************************************
       LIBNOTEMIDI - API for Linux notebook serial port MIDI

	(C) 2000 by Michael Minn - nycminn@earthlink.net

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

***************************************************************************/

#ifndef LIBNOTEMIDI_H
#define LIBNOTEMIDI_H

#include <string.h> // for memset

#define LIBNOTEMIDI_VERSION "0.2000.09.17 (c)2000 by Michael Minn"

#define SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"
#define NOTEMIDI_MAIN_PORT

class libnotemidi_file;

class libnotemidi
{
	public:
	libnotemidi(char *notemidi_port);
	~libnotemidi();
	int read_midi_file(char *midi_file_name, libnotemidi_file &midi_file);
	int play_midi_file(char *midi_file_name);
	int record_midi_file(char *midi_file_name);
	int is_ok() { return file_descriptor >= 0; }
	
	private:
	int file_descriptor;	
};

/*********************** MIDI Commands ****************/

#define LIBNOTEMIDI_NOTE_OFF 0x80
#define LIBNOTEMIDI_NOTE_ON  0x90
#define LIBNOTEMIDI_POLYPHONIC_KEY_PRESSURE 0xA0
#define LIBNOTEMIDI_CONTROL_CHANGE 0xB0
#define LIBNOTEMIDI_PROGRAM_CHANGE 0xC0
#define LIBNOTEMIDI_AFTERTOUCH 0xD0
#define LIBNOTEMIDI_PITCH_BEND 0xE0
#define LIBNOTEMIDI_SYSTEM_EXCLUSIVE 0xF0
#define LIBNOTEMIDI_TIME_CODE_QUARTER_FRAME 0xF1
#define LIBNOTEMIDI_SONG_POSITION 0xF2
#define LIBNOTEMIDI_SONG_SELECT 0xF3
#define LIBNOTEMIDI_TUNE_REQUEST 0xF6
#define LIBNOTEMIDI_END_OF_EXCLUSIVE 0xF7
#define LIBNOTEMIDI_TIMING_CLOCK 0xF8
#define LIBNOTEMIDI_START 0xFA
#define LIBNOTEMIDI_CONTINUE 0xFB
#define LIBNOTEMIDI_STOP 0xFC
#define LIBNOTEMIDI_ACTIVE_SENSING 0xFE
#define LIBNOTEMIDI_META_EVENT 0xFF

#define LIBNOTEMIDI_CONTROLLER_BANK_SELECT_MSB 0x00
#define LIBNOTEMIDI_CONTROLLER_MODULATION 0x01
#define LIBNOTEMIDI_CONTROLLER_BREATH_CONTROL 0x02
#define LIBNOTEMIDI_CONTROLLER_FOOT_CONTROL 0x04
#define LIBNOTEMIDI_CONTROLLER_PORTAMENTO_TIME 0x05
#define LIBNOTEMIDI_CONTROLLER_DATA_ENTRY_MSB 0x06
#define LIBNOTEMIDI_CONTROLLER_MAIN_VOLUME 0x07
#define LIBNOTEMIDI_CONTROLLER_BALANCE 0x08
#define LIBNOTEMIDI_CONTROLLER_PAN 0x0a
#define LIBNOTEMIDI_CONTROLLER_EXPRESSION 0x0b
#define LIBNOTEMIDI_CONTROLLER_EFFECT_1 0x0c
#define LIBNOTEMIDI_CONTROLLER_EFFECT_2 0x0d
#define LIBNOTEMIDI_CONTROLLER_BANK_SELECT_LSB 0x20
#define LIBNOTEMIDI_CONTROLLER_DATA_ENTRY_LSB 0x26
#define LIBNOTEMIDI_CONTROLLER_DAMPER 0x40
#define LIBNOTEMIDI_CONTROLLER_PORTAMENTO 0x41
#define LIBNOTEMIDI_CONTROLLER_SOSTENUTO 0x42
#define LIBNOTEMIDI_CONTROLLER_LEGATO 0x44
#define LIBNOTEMIDI_CONTROLLER_HARMONIC_CONTENT 0x47
#define LIBNOTEMIDI_CONTROLLER_RELEASE_TIME 0x48
#define LIBNOTEMIDI_CONTROLLER_ATTACK_TIME 0x49
#define LIBNOTEMIDI_CONTROLLER_BRIGHTNESS 0x4a
#define LIBNOTEMIDI_CONTROLLER_PORTAMENTO_CONTROLLER 0x54
#define LIBNOTEMIDI_CONTROLLER_REVERB_SEND 0x5b
#define LIBNOTEMIDI_CONTROLLER_CHORUS_SEND 0x5d
#define LIBNOTEMIDI_CONTROLLER_VARIATION_SEND 0x5e
#define LIBNOTEMIDI_CONTROLLER_NON_REG_LSB 0x62
#define LIBNOTEMIDI_CONTROLLER_NON_REG_MSB 0x63
#define LIBNOTEMIDI_CONTROLLER_REGISTERED_LSB 0x64
#define LIBNOTEMIDI_CONTROLLER_REGISTERED_MSB 0x65
#define LIBNOTEMIDI_CONTROLLER_ALL_SOUND_OFF 0x78
#define LIBNOTEMIDI_CONTROLLER_RESET_ALL 0x79
#define LIBNOTEMIDI_CONTROLLER_LOCAL_CONTROL 0x7a
#define LIBNOTEMIDI_CONTROLLER_ALL_NOTES_OFF 0x7b
#define LIBNOTEMIDI_CONTROLLER_OMNI_OFF 0x7c
#define LIBNOTEMIDI_CONTROLLER_OMNI_ON 0x7d
#define LIBNOTEMIDI_CONTROLLER_MONO_ON 0x7e
#define LIBNOTEMIDI_CONTROLLER_POLY_ON 0x7f

#define LIBNOTEMIDI_META_SEQUENCE 0x00
#define LIBNOTEMIDI_META_TEXT 0x01
#define LIBNOTEMIDI_META_COPYRIGHT 0x02
#define LIBNOTEMIDI_META_SEQUENCE_NAME 0x03
#define LIBNOTEMIDI_META_INSTRUMENT_NAME 0x04
#define LIBNOTEMIDI_META_LYRIC 0x05
#define LIBNOTEMIDI_META_MARKER 0x06
#define LIBNOTEMIDI_META_CUE_POINT 0x07
#define LIBNOTEMIDI_META_CHANNEL_PREFIX 0x20
#define LIBNOTEMIDI_META_END_OF_TRACK 0x2f
#define LIBNOTEMIDI_META_TEMPO 0x51
#define LIBNOTEMIDI_META_TIME_SIGNATURE 0x58
#define LIBNOTEMIDI_META_KEY_SIGNATURE 0x59
#define LIBNOTEMIDI_META_SEQUENCER_SPECIFIC 0x7f


/*********************** MIDI File Headers ****************/

struct libnotemidi_character_file_header
{
	char chunk_type[4];
	unsigned char length[4];
	unsigned char format[2];
	unsigned char tracks[2];
	unsigned char division[2];
};

struct libnotemidi_file_header
{
	char chunk_type[5];
	unsigned long length;
	unsigned long format;
	unsigned long tracks;
	unsigned long division;
	// unsigned long file_start_delay;
	libnotemidi_file_header()
	{
		strcpy(chunk_type, "MThd");
		length = 6;
		format = 0;
		tracks = 1;
		division = 480; // MIDI clock ticks per quarter note
		// file_start_delay = 480; // delay time at start of file to permit tempo change to settle
		// file_start_delay = 0; // don't do for now to test midi import
	}
};

struct libnotemidi_character_track_header
{
	char chunk_type[4];
	unsigned char length[4];
};

struct libnotemidi_track_header
{
	char chunk_type[5];
	unsigned long length;
	libnotemidi_track_header()
	{
		memset(this, (char)NULL, sizeof(libnotemidi_track_header));
		strcpy(chunk_type, "MTrk");
	}
};


/***************** Events ********************/

class libnotemidi_event
{
	public:
	int delta_time;
	int absolute_time;
	int absolute_milliseconds;
	libnotemidi_event *next;
	libnotemidi_event *previous;

	//public:
	unsigned char *data;

	libnotemidi_event() { data = NULL; next = NULL; previous = NULL; absolute_time = 0; }
	~libnotemidi_event() { if (data != NULL) delete[] data; data = NULL; }

	int set_absolute_times(int ticks_per_quarter_note);
	int find_tempo_beats_per_minute();
	int find_length();
	void delete_list();
	char *find_name(char *string);
	char *find_meta_event_name(char *string);
	libnotemidi_event *find_first_event();
	libnotemidi_event *add_event(libnotemidi_event *event);
	int find_number_of_delta_time_bytes() { if (delta_time <= 0x7f) return 1; else if (delta_time <= 0x3fff) return 2; else if (delta_time <= 0xfffff) return 3; else return 4; }
	int is_tempo() { return (this && data && (data[0] == LIBNOTEMIDI_META_EVENT) && (data[1] == LIBNOTEMIDI_META_TEMPO)); }

	unsigned char *write_delta_time(unsigned char *buffer);
	unsigned char *write_event(unsigned char *buffer);
	unsigned char *save_events(unsigned char *buffer);

	unsigned char *read_delta_time(unsigned char *buffer);
	unsigned char *read_event(unsigned char *buffer, unsigned char *buffer_end, unsigned char &current_status_byte);
	libnotemidi_event *load_events(unsigned char *(&buffer), int buffer_length, int &track_start_time);

	// Math Functions
	int long_from_bytes(int byte_count, unsigned char *bytes);
	void long_to_bytes(int byte_count, unsigned char *bytes, int value);
	int long_from_14bit(unsigned char *bytes)
	{
		int value = bytes[1] & 0x7f;
		value <<= 7;
		value += (bytes[0] & 0x7f);
		return value;
	}
	void long_to_14bit(int value, unsigned char *bytes)
	{
		bytes[0] = (unsigned char)(value & 0x7f);
		bytes[1] = (unsigned char)((value >> 7) & 0x7f);
	}
};

class libnotemidi_file
{
	public:
	libnotemidi_file() { events = NULL; }
	~libnotemidi_file () { events->delete_list(); events = NULL; }
	int save(unsigned char *buffer);
	unsigned char *load(unsigned char *buffer, int buffer_length);
	int find_division() { return file_header.division; }
	
	libnotemidi_event *events; // list_head

	private:
	libnotemidi_file_header file_header;
	libnotemidi_track_header track_header;

	unsigned char *read_file_header(unsigned char *buffer);
	unsigned char *read_track_header(unsigned char *buffer);

	unsigned char *write_file_header(unsigned char *buffer);
	unsigned char *write_track_header(unsigned char *buffer, int track_length);
	unsigned char *write_end_of_track(unsigned char *buffer);
};



#endif
